#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include "car.h"
#include "client.h"

using namespace std;

typedef std::vector<Renting*>     RentingVector;
typedef std::map<string, Car*>    CarMap;
typedef std::map<string, Client*> ClientMap;

typedef CarMap::value_type        CarMapValue;
typedef ClientMap::value_type     ClientMapValue;

//------------

Car*     selectCar          (CarMap&        cars);
Client*  selectClient       (ClientMap&     clients);
int      selectRentingIndex (RentingVector& clients);

void printCarList      (CarMap&   cars);
void showCarDetails    (CarMap&   cars);
void addNewCar         (CarMap&   cars);

void printClientList   (ClientMap& clients);
void showClientDetails (ClientMap& clients);
void addNewClient      (ClientMap& clients);

void rentCar           (CarMap& fl, RentingVector& sl, ClientMap& cl);
void removeRenting     (CarMap& fl, RentingVector& sl, ClientMap& cl);
void showProfit        (RentingVector& rentings);

void saveAll   (const string& fileName, 
                CarMap& fl, RentingVector& sl, ClientMap& cl);
void loadAll   (const string& fileName, 
                CarMap& fl, RentingVector& sl, ClientMap& cl);
void clearAll  (CarMap& fl, RentingVector& sl, ClientMap& cl);

//main                                                

int main ()
{
   const string CARS_FILE_NAME = "cars.txt";

   CarMap         cars;
   RentingVector  rentings;
   ClientMap      clients;
   
   loadAll(CARS_FILE_NAME, cars, rentings, clients);
   
   for (;;)
   {
      cout << "|-----------------------------|" << endl
           << "|----[   MASINU NUOMA:   ]----|" << endl
           << "|-----------------------------|" << endl
           << "|--  1. Sukurti masina      --|" << endl
           << "|--  2. Masinu sarasas      --|" << endl
           << "|--  3. Masinos detales     --|" << endl
           << "|-----------------------------|" << endl
           << "|--  4. Sukurti klienta     --|" << endl
           << "|--  5. Klientu sarasas     --|" << endl
           << "|--  6. Kliento detales     --|" << endl
           << "|-----------------------------|" << endl
           << "|--  7. Isnuomoti           --|" << endl
           << "|--  8. Panaikinti sutarti  --|" << endl
           << "|--  9. Laikotarpio pelnas  --|" << endl
           << "|-----------------------------|" << endl
           << "|--  0. Baigti darba        --|" << endl
           << "|-----------------------------|" << endl;
       string key;    
       cin >> key;
            if (key == "0") break;
       else if (key == "1") addNewCar(cars);
       else if (key == "2") printCarList(cars);
       else if (key == "3") showCarDetails(cars);
       else if (key == "4") addNewClient(clients);
       else if (key == "5") printClientList(clients);
       else if (key == "6") showClientDetails(clients);
       else if (key == "7") rentCar(cars, rentings, clients);
       else if (key == "8") removeRenting(cars, rentings, clients);
       else if (key == "9") showProfit(rentings);
       else cout << endl << "Neteisinga komanda..." << endl;   
   }
   
   saveAll(CARS_FILE_NAME, cars, rentings, clients);
   
   clearAll(cars, rentings, clients);
}

//---------

Car* selectCar (CarMap& cars)
{
   if (cars.empty())
   {
      cout << "Masinu sarasas yra tuscias." << endl;
      return NULL;
   }
   for (;;)
   {
      string name;
      printCarList(cars);
      cout << "Pasirinkite masina ('x'-nutraukti): ";
      cin >> name;
      if (name == "x")
         return NULL;
      if (cars.find(name) != cars.end())
      return cars[name];
   }
   return NULL; // unreacheable
}

//---------

Client* selectClient (ClientMap& clients)
{
   if (clients.empty())
   {
      cout << "Klientu sarasas yra tuscias." << endl;
      return NULL;
   }
   for (;;)
   {
      string name;
      printClientList(clients);
      cout << "Pasirinkite klienta ('x'-nutraukti): ";
      cin >> name;
      if (name == "x")
          return NULL;
      if (clients.find(name) != clients.end())
          return clients[name];
   }
   return NULL; // unreacheable
}

//----------

int selectRentingIndex (RentingVector& rentings)
{
   if (rentings.empty())
   {
      cout << "Nuomos sutarciu sarasas yra tuscias." << endl;
      return -1;
   }
   for (;;)
   {
      string dateFrom;
      string dateTo;
      cout << "--[ Pasirinkite nuomos laikotarpi ]--" << endl;
      cout << "   nuo kada (yyyy.mm.dd) '0' - nuo pradzios: "; cin >> dateFrom;
      cout << "   iki kada (yyyy.mm.dd) '9' - iki pabaigos: "; cin >> dateTo;
      for (int i = 0;  i < rentings.size();  i++)
         if (dateFrom               <= rentings[i]->getDate()  &&
            rentings[i]->getDate() <= dateTo)
         {
            cout << (i+1) << ".\t";
            rentings[i]->print();
         }
      string index;
      cout << "Pasirinkite pardavimo nr ('x'-nutraukti, '0'-keisti laikotarpi): ";
      cin >> index;
      if (index == "x")
          return -1;
      int s = 0;
      s = atoi(index.c_str());    
      if (s <= 0  ||  s > rentings.size())
          continue;
      return s - 1;
   }
   return -1; // unreacheable
}

//----------

void printCarList (CarMap& cars)
{
   cout << "--[ Masinu sarasas ]--[" << cars.size() << "]--" <<  endl;
   CarMap::iterator iter = cars.begin();
   for (;  iter != cars.end();  iter++)
      iter->second->print();
}

//----------

void printClientList (ClientMap& clients)
{
   cout << "--[ Klientu sarasas ]--" << clients.size() << endl;
   ClientMap::iterator iter = clients.begin();
   for (;  iter != clients.end();  iter++)
      iter->second->print();
}

//------------

void addNewCar (CarMap& cars)
{
   string name;
   for (;;)
   {
      printCarList(cars);
      cout << "--[ Nauja masina ]--" << endl
           << "Naujos masinos pavadinimas ('x'-nutraukti): " << endl;
      cin >> name;
      if (name == "x")
          return;
      if (cars.find(name) == cars.end())
          break;
      cout << "KLAIDA: tokia masina jau yra sarase!" << endl;    
   }
   double price;
   cout << "Dienos kaina: ";        
   cin  >> price;
   cars.insert(CarMapValue(name, new Car(name, price)));
}

//------------

void addNewClient (ClientMap& clients)
{
   string name;
   for (;;)
   {
      printClientList(clients);
      cout << "--[ Naujas klientas ]--" << endl
           << "Naujo kliento vardas ('x'-nutraukti): " << endl;
      cin >> name;
      if (name == "x")
          return;
      if (clients.find(name) == clients.end())
          break;
      cout << "KLAIDA: toks klientas jau yra sarase!" << endl;    
   }
   clients.insert(ClientMapValue(name, new Client(name)));
}

//-------------

void showCarDetails (CarMap& cars)
{
   cout << "--[ Masinos nuomos detales ]--" << endl;
   Car* car = selectCar(cars);
   if (car == NULL) 
      return;
   cout << "Masinos nuomos sutartys: ";
   car->print();
   for (int i = 0;  i < car->getRentingCount();  i++)
      car->getRenting(i)->print();
}

//------------

void showClientDetails (ClientMap& clients)
{
   cout << "--[ Kliento nuomos detales ]--" << endl;
   Client* client = selectClient(clients);
   if (client == NULL) 
      return;
   cout << "Kliento nuomos sutartys: ";
   client->print();
   for (int i = 0;  i < client->getRentingCount();  i++)
      client->getRenting(i)->print();
}

//--------------

void rentCar  (CarMap&         cars, 
               RentingVector&  rentings, 
               ClientMap&      clients)
{
   cout << "--[ Masinos isnuomavimas ]--" << endl;
   Car* car = selectCar(cars);
   if (car == NULL) 
      return;
   Client* client = selectClient(clients);
   if (client == NULL) 
      return;
   string date;
   int    dayCount;
   //double itemPrice;
   car->print();
   cout << "Nuomos data (yyyy.mm.dd): ";  cin >> date;
   cout << "Dienu skaicius: ";            cin >> dayCount;
   //cout << "vieneto kaina: ";               cin >> itemPrice;
   Renting* renting =
      new Renting(car, client, date, dayCount/*, itemPrice*/);
   car  ->addRenting(renting);
   client->addRenting(renting);
   rentings.push_back(renting);
}

void removeRenting (CarMap&         cars, 
                    RentingVector&  rentings, 
                    ClientMap&      clients)
{
   cout << "--[ Sutarties panaikinimas ]--" << endl;
   int i = selectRentingIndex(rentings);
   if (i < 0) 
       return;
   Renting* renting = rentings[i];    
   // cast from const Food* to call non-const removeRenting
   Car*     car     = (Car*) renting->getCar(); 
   car->removeRenting(renting);
   rentings.erase(rentings.begin() + i);
}
//------------------------------------------------------------

void showProfit (RentingVector& rentings)
{
   string dateFrom;
   string dateTo;
   cout << "--[ Laikotarpio pelnas ]--" << endl;
   cout << "   nuo kada (yyyy.mm.dd) '0' - nuo pradzios: "; cin >> dateFrom;
   cout << "   iki kada (yyyy.mm.dd) '9' - iki pabaigos: "; cin >> dateTo;
   double totalProfit = 0;
   for (int i = 0;  i < rentings.size();  i++)
      if (dateFrom <= rentings[i]->getDate()  &&
          dateTo   >= rentings[i]->getDate())
      {
         rentings[i]->print();
         totalProfit += rentings[i]->getProfit();
      }
   cout << "Laikotarpio pelnas: " << totalProfit << endl;
}

//-----------

void saveAll   (const string&   fileName, 
                CarMap&         cars, 
                RentingVector&  rentings, 
                ClientMap&      clients)
{
   ofstream out (fileName.c_str());
   
   out << cars.size() << endl;
   CarMap::iterator fi = cars.begin();
   for (;  fi != cars.end();  fi++)
      fi->second->save(out);
      
   out << clients.size() << endl;
   ClientMap::iterator ci = clients.begin();
   for (;  ci != clients.end();  ci++)
       ci->second->save(out);
      
   out << rentings.size() << endl;
   for (int i = 0;  i < rentings.size();  i++)
       rentings[i]->save(out);
       
   out.close();    
}                

//-----------

void loadAll   (const string&   fileName, 
                CarMap&         cars, 
                RentingVector&  rentings, 
                ClientMap&      clients)
{
   clearAll(cars, rentings, clients);

   ifstream  in (fileName.c_str());
   
   int carCount = 0;
   in >> carCount;
   for (int i = 0;  i < carCount;  i++)
   {
       Car* car = new Car(in);
       cars.insert(CarMapValue(car->getName(), car));
   }
       
   int clientCount = 0;
   in >> clientCount;
   for (int i = 0;  i < clientCount;  i++)
   {
       Client* client = new Client(in);
       clients.insert(ClientMapValue(client->getName(), client));
   }

   int rentingCount = 0;
   in >> rentingCount;
   for (int i = 0;  i < rentingCount;  i++)
       rentings.push_back(new Renting(in, cars, clients));
       
   in.close();    
}

//------------

void clearAll (CarMap&         cars, 
               RentingVector&  rentings, 
               ClientMap&      clients)
{
   CarMap::iterator fi = cars.begin();
   for (;  fi != cars.end();  fi++)
      delete fi->second;
   cars.clear();   

   ClientMap::iterator ci = clients.begin();
   for (;  ci != clients.end();  ci++)
      delete ci->second;
   clients.clear();   

   for (int i = 0;  i < rentings.size();  i++)
       delete rentings[i];
   rentings.clear();    
}

